@@ -69,7 +69,14 @@ class OpenGL(object):
         elif "intel" in vendor_string:
             self._gpu_vendor = OpenGL.Vendor.Intel
 
-        self._gpu_type = self._gl.glGetString(self._gl.GL_RENDERER)
+        #WORKAROUND: Cura/#1117 Cura-packaging/12
+        # Some Intel GPU chipsets return a string, which is not undecodable via PyQt5.
+        # This workaround makes the code fall back to a "Unknown" renderer in these cases.
+        try:
+            self._gpu_type = self._gl.glGetString(self._gl.GL_RENDERER)
+        except UnicodeDecodeError:
+            Logger.log("e", "DecodeError while getting GL_RENDERER via glGetString!")
+            self._gpu_type = "Unknown"
 
         if not self.hasFrameBufferObjects():
             Logger.log("w", "No frame buffer support, falling back to texture copies.")
@@ -248,7 +255,7 @@ class OpenGL(object):
     #
     #   \return The singleton instance.
     @classmethod
-    def getInstance(cls):
+    def getInstance(cls) -> "OpenGL":
         return cls._instance
 
     ##  Set the singleton instance.
